// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#pragma once

#include <utility>

#include "BlockablePurpose.hpp"
#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Context.hpp"
#include "ContextManager.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "lib/adt/LockedSet.hpp"
#include "stats/FromAnywhere.hpp"
#include "stats/Worker.hpp"

static emper::lib::adt::LockedSet<Context*> blockedContexts;

template <LogSubsystem logSubsystem>
class Blockable : public Logger<logSubsystem> {
 protected:
	Runtime& runtime;

	ContextManager& contextManager;

	const emper::BlockablePurpose blockablePurpose;

	workeraffinity_t* affinity = nullptr;

	// It would be OK to not initialize this member since it is only set
	// and read after affinity has been set to a non nullptr
	// value. However, if we do not set it, then clang-tidy complains
	// about clang-analyzer-optin.cplusplus.UninitializedObject.
	workeraffinity_t affinity_buffer = Fiber::NOT_AFFINE;

	Blockable(Runtime& runtime,
						emper::BlockablePurpose blockablePurpose = emper::BlockablePurpose::GENERIC)
			: runtime(runtime),
				contextManager(runtime.getContextManager()),
				blockablePurpose(blockablePurpose) {}

	void maybeSetAffinity() {
		if constexpr (!emper::SET_AFFINITY_ON_BLOCK && !emper::STATS) return;

		// TODO: At some point we may want to have something like
		// Runtime::isAffinityCapable() and return here if it is
		// not. Although I am not sure if it is worth the overhead.

		bool setAffinityBuffer = false;
		auto* currentFiber = Context::getCurrentFiber();
		auto* affinity = currentFiber->getAffinityBuffer();
		if (affinity) {
			this->affinity = affinity;
		} else {
			setAffinityBuffer = true;
			this->affinity = &affinity_buffer;
		}

		if (emper::STATS || setAffinityBuffer) {
			affinity_buffer = Runtime::getWorkerId();
		}
	}

	// Older clang-tidy versions show a
	// performance-unnecessary-value-param, newer versions do not show
	// this error if std::move() is used (as we do).
	// NOLINTNEXTLINE(performance-unnecessary-value-param)
	void block(func_t freshContextHook) {
		// Only contexts managed by EMPER's runtime can block.
		emper::assertInRuntime();

		LOGD("block() blockedContext=" << Context::getCurrentContext());

		maybeSetAffinity();

		decltype(blockedContexts)::set_type::size_type blockedContextCount = 0;
		if constexpr (emper::BLOCKED_CONTEXT_SET) {
			auto* currentContext = Context::getCurrentContext();

			if constexpr (emper::STATS) {
				blockedContextCount = blockedContexts.insertAndGetSize(currentContext);
			} else {
				blockedContexts.insert(currentContext);
			}
		}

		if constexpr (emper::STATS) {
			// Note that blockedContextCount may not been set here if
			// emper::BLOCKED_CONTEXT_SET is not enabled. But we will at
			// least record a blocked context here if emper::STATS is
			// enabled.
			emper::stats::worker->recordBlockedContext(blockedContextCount, blockablePurpose);
		}

		contextManager.saveAndStartNew(std::move(freshContextHook));
	}

	auto unblockAndGetContinuation(Context* context) -> Fiber* {
		assert(context != nullptr);
		if constexpr (emper::BLOCKED_CONTEXT_SET) {
			blockedContexts.erase(context);
		}

		if constexpr (emper::STATS) {
			if (likely(emper::stats::worker)) {
				emper::stats::worker->recordUnblockedContext(affinity_buffer, blockablePurpose);
			} else {
				// The unblock was triggered from outside the runtime system,
				// e.g. the Completer calling signal() on a semaphore of a IO
				// future.
				runtime.fromAnywhereStats->recordUnblockedContext(affinity_buffer, blockablePurpose);
			}
		}

		return Fiber::from([this, context]() { contextManager.discardAndResume(context); }, affinity);
	}

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void unblock(Context* context) {
		Fiber* unblockFiber = unblockAndGetContinuation(context);

		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			runtime.schedule(*unblockFiber);
		} else {
			runtime.scheduleFromAnywhere(*unblockFiber);
		}
	}
};
