// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <cassert>
#include <cstddef>
#include <iostream>
#include <string>

#include "Worker.hpp"
#include "emper-config.h"

namespace emper {

struct Emper {
	static auto inRuntime() -> bool { return Worker::isWorkerThread(); }
	static void assertInRuntime() { assert(inRuntime()); }
};

constexpr auto assertInRuntime = Emper::assertInRuntime;
constexpr auto inRuntime = Emper::inRuntime;

void printInfo(std::ostream& strm = std::cout);

const size_t WS_VICTIM_COUNT = EMPER_WS_VICTIM_COUNT;
const size_t WS_VICTIM_DENOMINATOR = EMPER_WS_VICTIM_DENOMINATOR;

enum class StatsStackUsage {
	disabled,
	efficient,
	accurate,
};

const StatsStackUsage STATS_STACK_USAGE = StatsStackUsage::EMPER_STATS_STACK_USAGE;

const bool STATS_STACK_USAGE_ENABLED = STATS_STACK_USAGE != StatsStackUsage::disabled;

const bool STATS =
#ifdef EMPER_STATS
		true
#else
		false
#endif
		|| STATS_STACK_USAGE_ENABLED;

template <typename C>
void statsIncr(C& counter) {
	if constexpr (STATS) {
		counter++;
	}
}

template <typename C>
void statsAdd(C& counter, long a) {
	if constexpr (STATS) {
		counter += a;
	}
}

const bool WORKER_SLEEP =
#ifdef EMPER_WORKER_SLEEP
		true
#else
		false
#endif
		;

enum class WorkerWakeupStrategy {
	one,
	throttle,
	all,
};

const enum WorkerWakeupStrategy WORKER_WAKEUP_STRATEGY =
		WorkerWakeupStrategy::EMPER_WORKER_WAKEUP_STRATEGY;

const bool WORKER_IGNORE_WAKEUP_HINT =
#ifdef EMPER_WORKER_IGNORE_WAKEUP_HINT
		true
#else
		false
#endif
		;

const bool LIBURCU =
#ifdef EMPER_LIBURCU
		true
#else
		false
#endif
		;

const bool WAITFREE_WORK_STEALING =
#ifdef EMPER_WAITFREE_WORK_STEALING
		true
#else
		false
#endif
		;

const bool DEBUG =
#ifndef NDEBUG
		true
#else
		false
#endif
		;

enum class LogTimeStamp {
	none,
	utc,
	local,
};

const enum LogTimeStamp LOG_TIMESTAMP = LogTimeStamp::EMPER_LOG_TIMESTAMP;

const bool OVERFLOW_QUEUE =
#ifdef EMPER_OVERFLOW_QUEUE
		true
#else
		false
#endif
		;

const bool CHECK_ANYWHERE_QUEUE_WHILE_STEALING =
#ifdef EMPER_CHECK_ANYWHERE_QUEUE_WHILE_STEALING
		true
#else
		false
#endif
		;

auto getFullVersion() -> std::string;

const bool BLOCKED_CONTEXT_SET =
#ifdef EMPER_BLOCKED_CONTEXT_SET
		DEBUG
#else
		false
#endif
		;

const bool IO =
#ifdef EMPER_IO
		true
#else
		false
#endif
		;

const bool IO_STEALING =
#ifdef EMPER_IO_STEALING
		true
#else
		false
#endif
		;

const bool IO_LOCKLESS_CQ =
#ifdef EMPER_IO_LOCKLESS_CQ
		true
#else
		false
#endif
		;

enum class IoLocklessMemoryOrder {
	weak,
	strong,
};

const enum IoLocklessMemoryOrder IO_LOCKLESS_MEMORY_ORDER =
		IoLocklessMemoryOrder::EMPER_IO_LOCKLESS_MEMORY_ORDER;

const bool IO_SINGLE_URING =
#ifdef EMPER_IO_SINGLE_URING
		true
#else
		false
#endif
		;

const bool IO_TRY_SYSCALL =
#ifdef EMPER_IO_TRY_SYSCALL
		true
#else
		false
#endif
		;

const bool IO_WORKER_URING = IO && !IO_SINGLE_URING;

const bool WAITFREE_IO_STEALING =
#ifdef EMPER_IO_WAITFREE_STEALING
		true
#else
		false
#endif
		;
// Initialize this bool in Emper.cpp because it needs code evaluation
// (LinuxVersion::compare) during runtime.
// Using a static variable here means EACH object file including this header has to
// evaluate the needed code during library initialization.
// An extern variable results in a single execution during initialization of the
// Emper.cpp object.
// This also has the advantage that the probability we crash because printing
// warnings during the comparison use not yet initialized components is reduced.
extern const bool IO_MUST_INVALIDATE_BROKEN_CHAIN;

enum class IoSqPoller {
	off,
	one,
	each,
	numa,
};

const enum IoSqPoller IO_SQ_POLLER = IoSqPoller::EMPER_IO_SQ_POLLER;

enum class IoCompleterBehavior {
	schedule,
	wakeup,
	none,
};

const enum IoCompleterBehavior IO_COMPLETER_BEHAVIOR =
		IoCompleterBehavior::EMPER_IO_COMPLETER_BEHAVIOR;

const bool SET_AFFINITY_ON_BLOCK =
#ifdef EMPER_SET_AFFINITY_ON_BLOCK
		true
#else
		false
#endif
		;

const bool HAS_FS_PATH =
#ifdef EMPER_HAS_FS_PATH
		true
#else
		false
#endif
		;

enum class CompleterSchedParam {
	unchanged,
	normal,
	nice_10,
	nice_19,
	idle,
};

const CompleterSchedParam COMPLETER_SCHED_PARAM =
		CompleterSchedParam::EMPER_IO_COMPLETER_SCHED_PARAM;

enum class ContinuationStealingMode {
	disabled,
	locked,
	waitfree,
};

const ContinuationStealingMode CONTINUATION_STEALING_MODE =
		ContinuationStealingMode::EMPER_CONTINUATION_STEALING_MODE;

const bool CONTINUATION_STEALING_MODE_LOCKED =
		CONTINUATION_STEALING_MODE == ContinuationStealingMode::locked;

const bool CONTINUATION_STEALING = CONTINUATION_STEALING_MODE != ContinuationStealingMode::disabled;

enum class ContinuationStealingMadviseStack {
	no,
	dontneed,
	free,
};

const enum ContinuationStealingMadviseStack CONTINUATION_STEALING_MADVISE_STACK =
		ContinuationStealingMadviseStack::EMPER_CONTINUATION_STEALING_MADVISE_STACK;

const bool BUILD_WITH_CLANG =
#ifdef EMPER_BUILD_WITH_CLANG
		true
#else
		false
#endif
		;

const bool CONTEXT_MANAGER_WITH_MEMORY_MANAGER =
#ifdef EMPER_CONTEXT_MANAGER_WITH_MEMORY_MANAGER
		true
#else
		false
#endif
		;

constexpr float MEMORY_MANAGER_VICTIM_PERCENTAGE = EMPER_MEMORY_MANAGER_VICTIM_PERCENTAGE / 100;
static_assert(MEMORY_MANAGER_VICTIM_PERCENTAGE >= 0 && MEMORY_MANAGER_VICTIM_PERCENTAGE <= 1);

const size_t ASSUME_PAGE_SIZE = EMPER_ASSUME_PAGE_SIZE;

const size_t MIN_CONTEXT_STACK_SIZE = EMPER_MIN_CONTEXT_STACK_SIZE;

const size_t ASSUME_CACHE_LINE_SIZE = EMPER_ASSUME_CACHE_LINE_SIZE;

const bool STACK_GUARD_PAGE =
#ifdef EMPER_STACK_GUARD_PAGE
		true
#else
		false
#endif
		;

}	 // namespace emper
