// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Nicolas Pfeiffer, Florian Schmaus
#pragma once

#include <atomic>
#include <cassert>
#include <cstddef>
#include <cstdint>
#include <functional>
#include <mutex>
#include <ostream>
#include <type_traits>

#include "AbstractFiber.hpp"
#include "Common.hpp"
#include "Context.hpp"
#include "ContextManager.hpp"
#include "Continuation.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Scheduler.hpp"
#include "emper-config.h"

#define emper_fibril __attribute__((optimize("no-omit-frame-pointer")))

namespace emper {
void maybeLockFibril(void* queueItem);
}

namespace adt {
template <typename I, const uintptr_t SIZE>
class LockedQueue;	// IWYU pragma: keep
}

class Fibril : public AbstractFiber, public Logger<LogSubsystem::FIBRIL> {
 private:
	using activeChildrenCount_t = std::conditional<emper::CONTINUATION_STEALING_MODE ==
																										 emper::ContinuationStealingMode::waitfree,
																								 std::atomic<uint32_t>, int>::type;
	activeChildrenCount_t activeChildrenCount = 0;

	// Members for locked continuation stealing.
	std::mutex fibrilMutex;
	friend void emper::maybeLockFibril(void* queueItem);

	// Members for wait-free continuation stealing.

	uint32_t reverseStealCount = 0;

	// Only used if continuation_stealing_madvise_stack is not 'no'.
	std::atomic<bool> resumable = false;

	Continuation cont;

	Context* const initialContext;

	Runtime& runtime;

	static thread_local Fibril* toResume;

	static constexpr bool UNMAP_STACKS =
			emper::CONTINUATION_STEALING_MADVISE_STACK != emper::ContinuationStealingMadviseStack::no;
	static constexpr bool LOCKED_CONTINUATION_STEALING =
			emper::CONTINUATION_STEALING_MODE == emper::ContinuationStealingMode::locked;

	inline void tryResume(uint32_t fibrilResumeValue) {
		LOGD("tryResumeFiber() invoking tryResume(" << fibrilResumeValue << ")");

		if constexpr (LOCKED_CONTINUATION_STEALING) {
			tryResumeLocked();
		} else {
			tryResumeWaitFree(fibrilResumeValue);
		}
	}

	inline void tryResumeLocked() {
		fibrilMutex.lock();
		int c = --activeChildrenCount;
		LOGD("tryResume() c=" << c << " after decrementing activeChildrenCount");

		const bool syncConditionHolds = c == 0;
		Context* const currentContext = Context::getCurrentContext();

		if (syncConditionHolds) {
			fibrilMutex.unlock();

			if (initialContext != currentContext) {
				LOGD("tryResume: current context " << currentContext << " != initialContext "
																					 << initialContext << ", freeing current context");
				runtime.getContextManager().putFreeContext(currentContext);

				// After the cont.execute() below, the current context will be
				// the initial context.
				Context::currentContext = initialContext;
			}

			// Resume this fibril, does *not* return.
			cont.execute(cont.sp);
			__builtin_unreachable();
		}

		if (initialContext == currentContext) {
			if constexpr (UNMAP_STACKS) {
				// Unmap unused stack pages.
				currentContext->unmap(cont.sp);
			}

			// Set currrentContext to nullptr, as otherwise we would
			// recycle/reuse it in ContextManager::start(). But this is the
			// initial context of this Fibril, which we can not
			// recycle/reuse at this point.
			Context::currentContext = nullptr;
		}

		fibrilMutex.unlock();

		// Continue in the scheduler to find new work.
	}

	inline void tryResumeWaitFree(uint32_t fibrilResumeValue) {
		// We need to save a copy of the initial context value before
		// performing the fetch_sub() below. As after the fetch_sub()
		// performed below, this Fibril instance may become destructed. We
		// know that it will not be yet destructed if we are the
		// control-flow who founds that the sync condition holds, because
		// then we are the one who is destructing it. But if the sync
		// condition does not hold, then we can not assume that the Fibril
		// instance is still alive, because a concurrenct control flow may
		// have observed the sync condition and destructed it.
		Context* const savedInitialContext = initialContext;

		// Note that the memory order for the fetch_sub below is acq_rel,
		// to ensure that the 'initialContext' member value is loaded into
		// 'savedInitialContext' before the decrement. As right after the
		// decrement, this Fibril instance could become invalid (if the
		// sync condition holds).

		uint32_t c;
#ifdef EMPER_CONTINUATION_STEALING_MODE_WAITFREE
		c = activeChildrenCount.fetch_sub(fibrilResumeValue, std::memory_order_acq_rel) -
				fibrilResumeValue;
#else
		// "Dummy" implementation for the case where 'activeChildrenCount' is not an std::atomic.
		c = activeChildrenCount -= fibrilResumeValue;
#endif
		LOGD("tryResume() c=" << c << " after substracting " << fibrilResumeValue);

		const bool syncConditionHolds = c == 0;
		Context* const currentContext = Context::getCurrentContext();

		if (syncConditionHolds) {
			if (initialContext != currentContext) {
				if constexpr (UNMAP_STACKS) {
					// Check the atomic resumable bool to see if the context
					// owner is not unmapping in parallel here.
					if (!resumable.exchange(true, std::memory_order_acq_rel)) {
						// The context owner is unmapping pages. We do not wait
						// for it to complete, instead wo return to the
						// scheduler. The unmapping owner will see that we tried
						// to resume and will resume for us.
						return;
					}
					// Otherwise continue with the resume operation.
				}
				LOGD("tryResume: current context " << Context::currentContext << " != initialContext "
																					 << initialContext << ", freeing current context");
				runtime.getContextManager().putFreeContext(currentContext);

				// After the cont.execute() below, the current context will be
				// the initial context.
				Context::currentContext = initialContext;
			}

			// Resume this fibril, does *not* return.
			cont.execute(cont.sp);
			__builtin_unreachable();
		}

		// At this point, we can not make any assumptions about the
		// validty of this Fibril instance. It may be already
		// destrcuted.
		if (savedInitialContext == currentContext) {
			// Here we know this Fibril instance is alive, because we are
			// on the initial Context where the Fibril instance was
			// constructed on.
			if constexpr (UNMAP_STACKS) {
				// The initial context, which is also the current context, is
				// now blocked. Employ the "practical solution" to the Cactus
				// Stack problem, as per Yang et. al. in their SPAA '16 paper,
				// and unmap, via madvise() MADV_FREE or MADV_DONTNEED, the
				// unused pages of the context (i.e., stack). Note that the
				// Context and hence the Fibril is not resumable while the
				// unmap/madvise operation is in progress.
				currentContext->unmap(cont.sp);

				if (true == resumable.exchange(true, std::memory_order_acq_rel)) {
					// Someone was trying to resume this Fiber while we where
					// unmapping/madvising the stack. We will resume it for them
					// now.
					cont.execute(cont.sp);
					__builtin_unreachable();
				}
			}

			// Set currrentContext to nullptr, as otherwise we would
			// recycle/reuse it in ContextManager::start(). But this is the
			// initial context of this Fibril, which we can not
			// recycle/reuse at this point.
			Context::currentContext = nullptr;
		}

		// Continue in the scheduler to find new work.
	}

 public:
	inline __attribute__((always_inline)) Fibril(Runtime& runtime = *Runtime::getRuntime())
			: cont(), initialContext(Context::getCurrentContext()), runtime(runtime) {
		if constexpr (emper::CONTINUATION_STEALING_MODE != emper::ContinuationStealingMode::waitfree &&
									emper::CONTINUATION_STEALING_MODE != emper::ContinuationStealingMode::locked) {
			DIE_MSG("Fibril can only be used if continuation_stealing_mode is enabled");
		}
		if constexpr (emper::BUILD_WITH_CLANG) {
			DIE_MSG("Fibril can not be used when build with clang");
		}

		LOGD("constructed " << *this);
	}

	~Fibril() override { join(); }

	auto asFibrilIfPossible() -> Fibril* override { return this; }

	void printTo(std::ostream& strm, bool withPtr) const override;

 private:
	void run() override;

	inline static void tryResumeFibril(uint32_t fibrilResumeValue) {
		Fibril* fibrilToResume = toResume;
		if (!fibrilToResume) return;

		// Set toResume to nullptr, as otherwise, once we restart the
		// current context in ContextManager::start() we would try to
		// resume this Fibril again. Note that it is important that we
		// set it to nullptr before we invoke tryResume(), as
		// tryResume() may never returns.
		toResume = nullptr;

		fibrilToResume->tryResume(fibrilResumeValue);
	}
	// ContextManager::start() invokes tryResumeFibril(uint32_t), let's be friends.
	friend void ContextManager::start();

	__attribute__((noinline, hot, optimize(3))) void doJoin() {
		cont.ip = __builtin_return_address(0);

		toResume = this;

		const uintptr_t fibrilResumeValue = [&] {
			if constexpr (LOCKED_CONTINUATION_STEALING) {
				return 1;
			} else {
				return reverseStealCount;
			}
		}();
		ContextManager::tryResumeFibril(fibrilResumeValue);
	}

	void checkContinuationState() {
		auto* poppedFiber = runtime.scheduler.popBottom();
		if (poppedFiber == this) {
			// Nobody stole the continuation, simply return and execute the continatuion ourselves.
			LOGD("poppedFiber is continuation");
			return;
		}
		LOGD("poppedFiber=" << poppedFiber << " was not this continuation fibril");

		if (poppedFiber) {
			// TODO: Is this also safe with locked continuation stealing?

			// We popped an unrelated fiber, reschedule it.
			[[maybe_unused]] bool res = runtime.scheduler.pushBottom(*poppedFiber);
			// Since we where the onces who just poped this Fiber, there
			// should be always enough space in the work-stealing queue to
			// push it back.
			assert(res);
		}

		// If the continuation was stolen, or if we popped an unrelated
		// fiber, then try to resume this Fibril.
		toResume = this;
		ContextManager::tryResumeFibril(1);
		__builtin_unreachable();
	}

 public:
	template <class RET, class... PARs, class... ARGs>
	inline __attribute__((always_inline)) void fork(RET* ret, RET (*fun)(PARs...), ARGs... args) {
		auto fork_func = [](ARGs... args, Fibril * fr, RET * ret, RET(*fun)(PARs...))
				__attribute__((noinline, hot, optimize(3))) {
			fr->cont.ip = __builtin_return_address(0);
			fr->runtime.scheduler.pushBottom(*fr);
			*ret = fun(args...);
			fr->checkContinuationState();
		};
		membar(fork_func(args..., this, ret, fun));
	}

	template <class... PARs, class... ARGs>
	inline __attribute__((always_inline)) void fork(void (*fun)(PARs...), ARGs... args) {
		auto fork_func = [](ARGs... args, Fibril * fr, void (*fun)(PARs...))
				__attribute__((noinline, hot, optimize(3))) {
			fr->cont.ip = __builtin_return_address(0);
			fr->runtime.scheduler.pushBottom(*fr);
			fun(args...);
			fr->checkContinuationState();
		};
		membar(fork_func(args..., this, fun));
	}

	template <class T, class... ARGs>
	inline __attribute__((always_inline)) void fork(std::function<T> fun, ARGs... args) {
		auto fork_func = [](Fibril * fr, std::function<T> fun, ARGs... args)
				__attribute__((noinline, hot, optimize(3))) {
			fr->cont.ip = __builtin_return_address(0);
			fr->runtime.scheduler.pushBottom(*fr);
			fun(args...);
			fr->checkContinuationState();
		};
		membar(fork_func(this, fun, args...));
	}

	inline __attribute__((always_inline)) void join() {
		if constexpr (LOCKED_CONTINUATION_STEALING) {
			joinLocked();
		} else {
			joinWaitFree();
		}
	}

 private:
	inline __attribute__((always_inline)) void joinLocked() {
		if (activeChildrenCount == 0) return;

		membar(doJoin());
	}

	inline __attribute__((always_inline)) void joinWaitFree() {
		if (reverseStealCount == 0) return;

		membar(doJoin());

		reverseStealCount = 0;
		if constexpr (UNMAP_STACKS) {
			resumable.store(false, std::memory_order_relaxed);
		}
	}

	// Disallow heap allocations of Fibril by declaring those private.
	static void* operator new(std::size_t);
	static void* operator new[](std::size_t);

	template <typename, const uintptr_t>
	friend class adt::LockedQueue;
};
