// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus, Florian Fischer
#include "Runtime.hpp"

#include <numa.h>
#include <pthread.h>	// for pthread_t, pthread_attr_init

#include <cerrno>	 // for errno
// Non portable.
#include <sched.h>				// for cpu_set_t, CPU_SET, CPU_ZERO
#include <sys/sysinfo.h>	// for get_nprocs
#include <unistd.h>

#include <array>
#include <cstdlib>	// for rand, srand, abort
#include <cstring>
#include <fstream>	// IWYU pragma: keep
#include <iostream>
#include <memory>	 // for __shared_ptr_access, shared_ptr
#include <string>	 // for string
#include <thread>

#include "Common.hpp"	 // for DIE_MSG_ERRNO, DIE, DIE_MSG
#include "Context.hpp"
#include "ContextManager.hpp"	 // for ContextManager
#include "Debug.hpp"					 // for DBG, ABORT, LOGD, LOGE
#include "Emper.hpp"
#include "Fiber.hpp"	// for Fiber
#include "FiberSource.hpp"
#include "NextFiberResult.hpp"
#include "RuntimeStrategy.hpp"
#include "RuntimeStrategyStats.hpp"	 // for RuntimeStrategyStats
#include "WorkerLocalData.hpp"
#include "emper-config.h"
#include "io/GlobalIoContext.hpp"
#include "io/IoContext.hpp"
#include "io/Stats.hpp"
#include "lib/DebugUtil.hpp"
#include "lib/env.hpp"
#include "log/LogBuffer.hpp"
#include "stats/FromAnywhere.hpp"
#include "stats/Worker.hpp"

#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
#include "strategies/ws/WsStrategy.hpp"
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
#include "strategies/laws/LawsStrategy.hpp"
#else
#error "Unknown default scheduling strategy"
#endif

#ifdef EMPER_LIBURCU
#include <urcu.h>	 // for rcu_register_thread
#endif

#ifndef EMPER_HAS_GETTID
#include <syscall.h>
#define gettid() (syscall(SYS_gettid))
#endif

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;

Runtime::StrategyFactory Runtime::DEFAULT_STRATEGY_FACTORY =
#ifdef EMPER_DEFAULT_SCHEDULING_STRATEGY_WORK_STEALING
		WsStrategy::getFactory()
#elif defined EMPER_DEFAULT_SCHEDULING_STRATEGY_LOCALITY_AWARE_WORK_STEALING
		LawsStrategy::getFactory()
#else
#error "Unknown default scheduling strategy"
#endif
		;

using emper::io::GlobalIoContext;
using emper::io::IoContext;

bool Runtime::guardPageTouchCausesExitSuccess = false;
struct sigaction Runtime::oldact {};

Runtime::Runtime(workerid_t workerCount, const std::vector<NewWorkerHook>& newWorkerHooks,
								 bool pinWorkers, workerid_t pinningOffset, const StrategyFactory& strategyFactory,
								 unsigned int seed)
		: workerCount(workerCount),
			workerLatch(workerCount),
			firstWorkerThreadExitLatch(workerCount),
			secondWorkerThreadExitLatch(workerCount),
			strategy(strategyFactory(*this)),
			scheduler(strategy->getScheduler()),
			dispatcher(strategy->getDispatcher()),
			contextManager(*(new ContextManager(*this))),
			threads(new pthread_t[workerCount]),
			workers(new Worker*[workerCount]),
			ioContexts(emper::IO ? workerCount : 0),
			ioReadySem(0),
			randomEngine(seed),
			workerSleepStrategy(*this, workerCount),
			wakeupStrategy(workerCount) {
	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	const char* envValueLogFile = std::getenv("EMPER_LOG_FILE");
	if (envValueLogFile) {
		logBuffer = new emper::log::LogBuffer(envValueLogFile);
	}

	if constexpr (emper::STATS) {
		workerStats = new WorkerLocalData<emper::stats::Worker>(
				[workerCount] {
					emper::stats::worker = new emper::stats::Worker(workerCount);
					return emper::stats::worker;
				},
				*this);
		fromAnywhereStats = new emper::stats::FromAnywhere(workerCount);
	} else {
		workerStats = nullptr;
		fromAnywhereStats = nullptr;
	}

	// transfere newWorkerHooks
	for (const auto& f : newWorkerHooks) this->newWorkerHooks.push_back(f);

	// initialize the global IoContext if a completer is used
	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		// The global io_uring needs at least workerCount entries in its SQ because
		// for each worker's IoContext one eventfd read is prepared before the
		// globalCompleter is started and submits all previously prepared sqes.
		const unsigned entries = emper::IO_SINGLE_URING ? EMPER_IO_WORKER_URING_ENTRIES : workerCount;
		globalIo = new GlobalIoContext(*this, entries);
		globalIo->startCompleter();

		if constexpr (emper::STATS) {
			globalIo->stats.workerId = emper::io::Stats::COMPLETER_ID;
		}
	}

	// Check if the pinning settings are sound
	if (pinningOffset && !pinWorkers) {
		DIE_MSG("pinningOffset and not pinning workers are mutually exclusive");
	}

	// When sharing SQ poll threads the IoContext which the other attach to
	// must be initialized first
	if constexpr (emper::IO_SQ_POLLER == emper::IoSqPoller::numa) {
		if (numa_available() < 0) {
			DIE_MSG("numa support not available for numa based sq poller sharing");
		}

		struct bitmask* nodeCpus = numa_allocate_cpumask();
		if (!nodeCpus) {
			DIE_MSG_ERRNO("numa_allocate_cpumask failed");
		}

		for (unsigned node = 0; node < numa_all_nodes_ptr->size; ++node) {
			if (!numa_bitmask_isbitset(numa_all_nodes_ptr, node)) {
				continue;
			}

			int err = numa_node_to_cpus(static_cast<int>(node), nodeCpus);
			if (err) {
				DIE_MSG_ERRNO("numa_node_to_cpu failed");
			}

			for (unsigned cpu = 0; cpu < nodeCpus->size; ++cpu) {
				if (!numa_bitmask_isbitset(nodeCpus, cpu)) {
					continue;
				}

				workerid_t workerId = cpuToWorkerId(cpu);
				ioContexts[workerId] = new IoContext(*this);
				break;
			}
		}
	}

	if constexpr (emper::STACK_GUARD_PAGE) {
		void (*currentSigSegvHandler)(int, siginfo_t*, void*);
		{
			struct sigaction currentSigaction;
			if (sigaction(SIGSEGV, nullptr, &currentSigaction))
				DIE_MSG_ERRNO("getting the current signal action for SIGSEGV");

			currentSigSegvHandler = currentSigaction.sa_sigaction;
		}

		// Maybe another Runtime instance has already installed our
		// handler, do nothing if this is the case. Note that we would
		// have to create a lock for this whole section, to avoid the
		// Time-of-Check-to-Time-of-Install-Sighandler problem. But we do
		// do that for now.
		if (currentSigSegvHandler != &sigSegvHandler) {
			struct sigaction sa;
			sigemptyset(&sa.sa_mask);
			sa.sa_sigaction = &sigSegvHandler;
			sa.sa_flags = SA_RESTART | SA_SIGINFO | SA_ONSTACK;

			if (sigaction(SIGSEGV, &sa, &oldact)) DIE_MSG_ERRNO("setting SIGSEGV handler failed");
		}
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		if (pinWorkers) {
			// Start non portable.
			cpu_set_t cpuset;
			CPU_ZERO(&cpuset);
			CPU_SET((i + pinningOffset) % nprocs, &cpuset);

			errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
			if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
			// End non portable.
		}

		// This is not done by each thread individually because the randomEngine is
		// mutated and therefore produces data races.
		unsigned int seed = uniformIntDistribution(randomEngine);
		auto* worker = new Worker(i, workerCount, seed);
		workers[i] = worker;

		auto thread_function = [](void* voidWorkerPtr) -> void* {
#ifdef EMPER_LIBURCU
			rcu_register_thread();
#endif

			auto* worker = reinterpret_cast<Worker*>(voidWorkerPtr);
			return currentRuntime->workerLoop(worker);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, worker);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");

		// Note that Linux limits the length of the thread name to 16
		// bytes, incl. the terminating null byte. Hence we have to use a
		// hip name.
		std::string thread_name = "EMPER Wrkr #" + std::to_string(i);
		errno = pthread_setname_np(threads[i], thread_name.c_str());
		if (errno) DIE_MSG_ERRNO("pthread_setname_np() failed");
	}

	threadsRunning = true;

	const std::string stacktraceOnAbortsEnvVar("EMPER_STACKTRACE_ON_ABORTS");
	const char* envValueEnableStacktraces = std::getenv(stacktraceOnAbortsEnvVar.c_str());
	if (envValueEnableStacktraces && !std::strcmp(envValueEnableStacktraces, "true")) {
		DBG(stacktraceOnAbortsEnvVar << " set to 'true': Enabling stacktrace on abort");
		enableStacktraceOnAborts();
	}
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	if (threadsRunning) {
		initiateAndWaitUntilTermination();
	}

	// Print the stats after all worker threads have terminated and before we delete
	// objects bound to the runtime's lifetime
	if constexpr (emper::STATS) {
		const char* envValueStatsFile = std::getenv("EMPER_STATS_FILE");
		if (envValueStatsFile) {
			std::ofstream out(envValueStatsFile, std::ofstream::out);
			printStats(out);
		} else {
			printStats(std::cout);
		}
	}

	for (unsigned int i = 0; i < workerCount; ++i) {
		delete workers[i];
	}
	delete[] workers;
	delete[] threads;

	if constexpr (emper::IO) {
		if constexpr (emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
			// It is safer to destroy the globalIo before the worker IoContexts
			// because pointer to the worker IoContexts can outlife their objects
			// in the globalCompleter
			delete globalIo;
		}

		for (unsigned int i = 0; i < workerCount; ++i) {
			delete ioContexts[i];
		}
	}

	if constexpr (emper::STATS) {
		workerStats->forEach([](auto* stats) { delete stats; });
		delete workerStats;
		delete fromAnywhereStats;
	}

	DBG("Runtime " << this << " terminated");

	// TODO: Move this before the delete invocations?
	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}

	delete logBuffer;
}

auto Runtime::workerLoop(Worker* worker) -> void* {
	worker->setWorker();

	if constexpr (emper::IO_WORKER_URING) {
		if (!ioContexts[worker->workerId]) {
			ioContexts[worker->workerId] = new IoContext(*this);
		}

		// notify the globalCompleter that we have initialized our IoContext (registered our eventfd)
		ioContexts[worker->workerId]->setWorkerIo(worker);
		ioReadySem.notify();
	}

	LOGD("Worker loop started by thread " << gettid());

	// Alternative stack to use for the SIGSEGV signal handler we use to
	// check for stack gurad-page hits. We need an alternative stack
	// because if the guard page is it, the stack is already overflown
	// and thus the signal handler can not run on the already overflown
	// stack.
	constexpr size_t sigSegvAltStackSize = emper::STACK_GUARD_PAGE ? SIGSTKSZ : 0;
	std::array<char, sigSegvAltStackSize> sigSegvAltStack;
	if constexpr (emper::STACK_GUARD_PAGE) {
		stack_t ss = {sigSegvAltStack.data(), 0, sigSegvAltStack.size()};

		if (sigaltstack(&ss, nullptr)) {
			DIE_MSG_ERRNO("setting alternative signal stack failed");
		}
	}

	int oldType;
	// NOLINTNEXTLINE: cert-pos47-c
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	for (const auto& f : newWorkerHooks) f(worker->workerId);

	workerLatch.count_down_and_wait();

	contextManager.start();

	// Threads return here if ContextManager::switchToOriginalStack() is called.

	// Ensure that all worker threads exit "at the same" time. Otherwise
	// it would be possible that one thread is work-stealing,
	// potentially accessing a work stealing queue of an worker thread
	// that already exited, causing an invalid access.
	firstWorkerThreadExitLatch.count_down_and_wait();

	if (worker->getWorkerId() == 0) {
		// Obtain stats when all workers reached the first latch but before they exit.
		strategyLastStats = strategy->getStats();
		// Note that we do not need to do the same for the
		// RuntimeWorkerStats, as these are dynamically allocated and
		// still available once the worker threads terminate.
	}

	secondWorkerThreadExitLatch.count_down_and_wait();

	return nullptr;
}

auto Runtime::getDefaultWorkerCount() -> workerid_t {
	static const std::string workerCountEnvVarName = "EMPER_WORKER_COUNT";
	auto workerCountEnv = emper::lib::env::getUnsignedFromEnv<workerid_t>(workerCountEnvVarName);
	if (workerCountEnv) {
		auto value = workerCountEnv.value();
		DBG("Processed " << workerCountEnvVarName << "=" << value);
		return value;
	}

	// The CPU count reported by sysconf(_SC_NPROCESSORS_ONLN), sysconf(_SC_NPROCESSORS_CONF)
	// or std::thread::hardware_concurrency() may not match the CPU count available
	// to the emper process.
	// This can happen for example if we run EMPER in a container or qemu.o
	// GNU nproc(1) does report the correct CPU count using pthread_getaffinity_np
	// which fills a cpu_set with the CPUs the current process can be scheduled on
	// therefore we use this more precise CPU count if possible.
	// Code inspiration taken from:
	// https://github.com/coreutils/gnulib/blob/90e79512d8b385801218d6e9c4d88ff77186560b/lib/nproc.c#L206
	cpu_set_t set;
	if (pthread_getaffinity_np(pthread_self(), sizeof(set), &set) == 0) {
		unsigned long count = CPU_COUNT(&set);
		if (count > 0) {
			return count;
		}
	}

	return std::thread::hardware_concurrency();
}

void Runtime::yield() {
	Context* context = Context::getCurrentContext();
	contextManager.saveAndStartNew([context, this] {
		Fiber* resumeFiber =
				Fiber::from([context, this] { this->contextManager.discardAndResume(context); });
		this->scheduleFromAnywhere(*resumeFiber);
	});
}

auto Runtime::nextFiber() -> std::optional<NextFiberResult> {
	if constexpr (emper::IO_WORKER_URING) {
		// Schedule all fibers waiting on completed IO
		IoContext::ContinuationBuffer completions;
		unsigned ncompletions =
				IoContext::getWorkerIo()->reapCompletions<CallerEnvironment::OWNER>(completions);
		if (ncompletions > 0) {
			// Keep the first and schedule the rest
			Fiber* next = completions[0];
			schedule(&completions[1], ncompletions - 1);

			return NextFiberResult{next, emper::FiberSource::io};
		}
	}

	return scheduler.nextFiber();
}

void Runtime::initiateTermination() {
	terminateWorkers.store(true, std::memory_order_release);

	workerSleepStrategy.notifyAll<CallerEnvironment::ANYWHERE>();

	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		globalIo->initiateTermination();
	}
}

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		errno = pthread_join(threads[i], nullptr);
		if (errno) {
			DIE_MSG_ERRNO("pthread_join() returned error for worker ID " << std::to_string(i));
		}
	}

	if constexpr (emper::IO && emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		globalIo->waitUntilFinished();
	}

	threadsRunning = false;
}

void Runtime::initiateAndWaitUntilTermination() {
	initiateTermination();
	waitUntilFinished();
}

void Runtime::printStats(std::ostream& out, bool detailed) {
	if constexpr (!emper::STATS) {
		out << "printStats(): EMPER stats disabled by static configuration" << std::endl;
		return;
	}

	std::shared_ptr<RuntimeStrategyStats> stats;
	if (terminateWorkers) {
		// The runtime was already requested to shut down. We can not
		// obtain stats as it is not guranteed that the worker threads are
		// still running in this case. We can however check
		// strategyLastStats and print that.

		if (strategyLastStats) {
			stats = strategyLastStats;
		} else {
			std::cerr << "printStats(): worker threads have already been scheduled for termination, but "
									 "not last stats recored yet."
								<< std::endl;
			return;
		}
	} else {
		stats = strategy->getStats();
	}

	stats->print(out);

	emper::stats::Worker globalWorkerStats(workerCount);
	auto workerStatsSnapshot = workerStats->getSnapshot();
	for (auto& stats : workerStatsSnapshot) {
		globalWorkerStats += stats;
	}

	globalWorkerStats.print(out);
	if (detailed) {
		for (auto& stats : workerStatsSnapshot) {
			stats.print();
		}
	}

	fromAnywhereStats->print(out);

	if constexpr (emper::IO) {
		emper::io::Stats::printStats(globalIo, ioContexts, out);
	}

	workerSleepStrategy.printStats(out);
}

void Runtime::printLastRuntimeStats(std::ostream& out) {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats(out);
}

void Runtime::executeAndWait(std::function<void()> f) {
	if (emper::inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	scheduleFromAnywhere(*fiber);

	fiberFinished.lock();
}

auto Runtime::shouldPinWorkers() -> bool {
	static const std::string pinWorkersEnvVarName = "EMPER_PIN_WORKERS";
	auto pinWorkersEnv = emper::lib::env::getBoolFromEnv(pinWorkersEnvVarName);
	if (pinWorkersEnv) {
		auto value = pinWorkersEnv.value();
		DBG("Processed " << pinWorkersEnvVarName << "=" << value);
		return value;
	}

	return true;
}

auto Runtime::getDefaultPinningOffset() -> workerid_t {
	static const std::string pinningOffsetEnvVarName = "EMPER_PINNING_OFFSET";
	auto pinningOffsetEnv = emper::lib::env::getUnsignedFromEnv<workerid_t>(pinningOffsetEnvVarName);
	if (pinningOffsetEnv) {
		auto value = pinningOffsetEnv.value();
		DBG("Processed " << pinningOffsetEnvVarName << "=" << value);
		return value;
	}

	return 0;
}

void Runtime::maybeTerminateWorker() {
	if (terminateWorkers.load(std::memory_order_relaxed)) {
		LOGD("Worker terminating");
		// Switch to the original stack, which will simply terminate the thread.
		ContextManager::switchToOriginalStack();
	}
}

void Runtime::sigSegvHandler(int sig, siginfo_t* info, ATTR_UNUSED void* ucontext) {
	if (sig != SIGSEGV) {
		static const char* const msg = "ERROR: Received unexpected signal";
		IGNORE_UNUSED_RESULT
		write(STDERR_FILENO, msg, strlen(msg));
		POP_DIAGNOSTIC
		exit(EXIT_FAILURE);
	}

	void* causingAddress = info->si_addr;

	if (Context::inCurrentGuardPage(causingAddress)) {
		static const char* const msg = "ERROR: SIGSEGV due to stack guard page\n";
		IGNORE_UNUSED_RESULT
		write(STDERR_FILENO, msg, strlen(msg));
		POP_DIAGNOSTIC
		if (guardPageTouchCausesExitSuccess) {
			exit(EXIT_SUCCESS);
		}
	}

	struct sigaction savedact;
	if (sigaction(sig, &oldact, &savedact))
		DIE_MSG_ERRNO("temporarily restoring original signal action for SIGSEGV");

	if (raise(sig)) DIE_MSG_ERRNO("re-raising SIGSEGV");

	// It is very unlikely that this is ever reached.
	if (sigaction(sig, &savedact, &oldact))
		DIE_MSG_ERRNO("temporarily restoring EMPER signal action for SIGSEGV");

	exit(EXIT_FAILURE);
}
