// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <utility>

#include "BinaryPrivateSemaphore.hpp"
#include "CountingPrivateSemaphore.hpp"
#include "Fiber.hpp"
#include "PrivateSemaphore.hpp"

class SynchronizedFiber : public Fiber {
 private:
	PrivateSemaphore& semaphore;

	SynchronizedFiber(fiber_fun_t function, void* arg, workeraffinity_t* affinity,
										PrivateSemaphore& semaphore)
			: Fiber(std::move(function), arg, affinity), semaphore(semaphore){};

	// cppcheck-suppress uninitMemberVar
	explicit SynchronizedFiber(const fiber_fun0_t& function, workeraffinity_t* affinity,
														 PrivateSemaphore& semaphore)
			: SynchronizedFiber([function](UNUSED_ARG void* arg) { function(); }, nullptr, affinity,
													semaphore) {}

	SynchronizedFiber(fiber_fun_t function, void* arg, PrivateSemaphore& semaphore)
			: SynchronizedFiber(std::move(function), arg, nullptr, semaphore){};

	// cppcheck-suppress uninitMemberVar
	explicit SynchronizedFiber(const fiber_fun0_t& function, PrivateSemaphore& semaphore)
			: SynchronizedFiber(function, nullptr, semaphore) {}

	void run() override {
		Fiber::run();
		semaphore.signalAndExit();
	}

 public:
	static inline auto from(fiber_fun_t function, void* arg, BinaryPrivateSemaphore& semaphore)
			-> Fiber* {
		return new SynchronizedFiber(std::move(function), arg, semaphore);
	}

	static inline auto from(const fiber_fun0_t& function, BinaryPrivateSemaphore& semaphore)
			-> Fiber* {
		return new SynchronizedFiber(function, semaphore);
	}

	static inline auto from(fiber_fun_t function, void* arg, workeraffinity_t* affinity,
													BinaryPrivateSemaphore& semaphore) -> Fiber* {
		return new SynchronizedFiber(std::move(function), arg, affinity, semaphore);
	}

	static inline auto from(const fiber_fun0_t& function, workeraffinity_t* affinity,
													BinaryPrivateSemaphore& semaphore) -> Fiber* {
		return new SynchronizedFiber(function, affinity, semaphore);
	}

	static inline auto from(fiber_fun_t function, void* arg, CountingPrivateSemaphore& semaphore)
			-> Fiber* {
		semaphore.incrementCounterByOne();
		return new SynchronizedFiber(std::move(function), arg, semaphore);
	}

	static inline auto from(const fiber_fun0_t& function, CountingPrivateSemaphore& semaphore)
			-> Fiber* {
		semaphore.incrementCounterByOne();
		return new SynchronizedFiber(function, semaphore);
	}

	static inline auto from(fiber_fun_t function, void* arg, workeraffinity_t* affinity,
													CountingPrivateSemaphore& semaphore) -> Fiber* {
		semaphore.incrementCounterByOne();
		return new SynchronizedFiber(std::move(function), arg, affinity, semaphore);
	}

	static inline auto from(const fiber_fun0_t& function, workeraffinity_t* affinity,
													CountingPrivateSemaphore& semaphore) -> Fiber* {
		semaphore.incrementCounterByOne();
		return new SynchronizedFiber(function, affinity, semaphore);
	}
};
