// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io.hpp"

#include <netdb.h>
#include <sys/socket.h>

#include <cerrno>
#include <cstring>
#include <filesystem>
#include <ostream>
#include <stdexcept>

#include "Common.hpp"
#include "CountingPrivateSemaphore.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "Fibril.hpp"
#include "Runtime.hpp"
#include "SynchronizedFiber.hpp"

namespace fs = std::filesystem;

namespace emper::io {

const int SockOpt::ENABLE = 1;
const SockOpt SockOpt::ReusePort = {SOL_SOCKET, SO_REUSEPORT, &SockOpt::ENABLE,
																		sizeof(SockOpt::ENABLE)};

static void recursive_directory_walk_helper(
		CPS& cps, const fs::path& dirpath,
		std::function<bool(const fs::directory_entry& dirent)>& filter,
		std::function<void(fs::directory_entry dirent)>& fn) {
	if constexpr (!emper::HAS_FS_PATH) {
		throw std::logic_error("recursive_directory_walk need c++17 std::filesystem::path");
	}

	Runtime* runtime = Runtime::getRuntime();
	for (const auto& p : fs::directory_iterator(dirpath)) {
		if (filter(p)) {
			auto* fnFiber = SynchronizedFiber::from([=] { fn(p); }, cps);
			runtime->schedule(*fnFiber);
		}

		if (p.is_directory()) {
			auto* subDirFiber = SynchronizedFiber::from(
					[&, p = p] { recursive_directory_walk_helper(cps, p, filter, fn); }, cps);
			runtime->schedule(*subDirFiber);
		}
	}
}

void recursive_directory_walk(const fs::path& dirpath,
															std::function<bool(const fs::directory_entry& dirent)> filter,
															std::function<void(fs::directory_entry dirent)> fn) {
	CPS cps;
	recursive_directory_walk_helper(cps, dirpath, filter, fn);
	cps.wait();
}

// NOLINTNEXTLINE(clang-diagnostic-unknown-attributes)
emper_fibril void fibril_recursive_directory_walk(
		// NOLINTNEXTLINE(performance-unnecessary-value-param)
		const fs::path dirpath, std::function<bool(const fs::directory_entry& dirent)> filter,
		// NOLINTNEXTLINE(performance-unnecessary-value-param)
		std::function<void(fs::directory_entry dirent)> fn) {
	Fibril fibril;

	for (const auto& pathRef : fs::directory_iterator(dirpath)) {
		// We must copy 'pathRef', to not hold a reference, which may
		// become invalid at the end of the iteration.
		const fs::directory_entry path = pathRef;
		if (filter(path)) {
			fibril.fork(fn, path);
		}

		if (pathRef.is_directory()) {
			fibril.fork(fibril_recursive_directory_walk, path, filter, fn);
		}
	}
}

auto tcp_listener(const std::string& host, const std::string& port,
									const std::function<void(int)>& handler, int backlog,
									const std::vector<SockOpt>& sockopts) -> Fiber* {
	int listen_socket;
	struct addrinfo* servaddr;

	Runtime* runtime = Runtime::getRuntime();

	int err = getaddrinfo(host.c_str(), port.c_str(), nullptr, &servaddr);
	if (err) {
		if (err == EAI_SYSTEM) {
			LOGE("getaddrinfo failed: " << strerror(errno));
			return nullptr;
		}

		LOGE("error in getaddrinfo: " << gai_strerror(err));
		return nullptr;
	}

	// socket creation and verification
	listen_socket = socket(servaddr->ai_family, servaddr->ai_socktype, servaddr->ai_protocol);
	if (listen_socket == -1) {
		LOGE("listen socket creation failed: " << strerror(errno));
		return nullptr;
	}

	for (const auto& sockopt : sockopts) {
		if (setsockopt(listen_socket, sockopt.level, sockopt.name, sockopt.val, sockopt.val_len) ==
				-1) {
			LOGE("setsockopt " << sockopt.name << " failed: " << strerror(errno));
			goto close_sock;
		}
	}

	if (bind(listen_socket, servaddr->ai_addr, servaddr->ai_addrlen) == -1) {
		LOGE("bind failed: " << strerror(errno));
		goto close_sock;
	}

	if (listen(listen_socket, backlog) != 0) {
		LOGE("listen failed: " << strerror(errno));
		goto close_sock;
	}

	return Fiber::from([=] {
		for (;;) {
			int client_fd = acceptAndWait(listen_socket, nullptr, nullptr);
			if (unlikely(client_fd < 0)) {
				LOGE("accept failed: " << strerror(errno));
				return;
			}

			Fiber* handler_fiber = Fiber::from([=] { handler(client_fd); });
			runtime->schedule(*handler_fiber);
		}
	});

close_sock:
	closeAndForget(listen_socket);
	return nullptr;
}
}	 // namespace emper::io
