// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <climits>
#include <cstdint>

#define TPTR_POINTER_BITS 48
#define TPTR_POINTER_MASK ((1UL << TPTR_POINTER_BITS) - 1)

static_assert(sizeof(uintptr_t) * CHAR_BIT - TPTR_POINTER_BITS == sizeof(uint16_t) * CHAR_BIT,
							"Tagged pointer assumptions are broken");

namespace emper::lib {
/**
 * @brief pointer which uses the 16 unused bits in x86-64 address space to store an uint16_t
 *
 * Additionally the least significant bit, which will likely be 0 because of alignment can be
 * used as a mark.
 */
class TaggedPtr {
 private:
	uintptr_t tptr = 0;

 public:
	template <typename T>
	TaggedPtr(T* ptr, uint16_t tag = 0, bool marked = false)
			: tptr(reinterpret_cast<uintptr_t>(ptr)) {
		setTag(tag);
		setMark(marked);
	}

	TaggedPtr(void* ptr) : tptr(reinterpret_cast<uintptr_t>(ptr)) {}
	TaggedPtr(uintptr_t ptr, uint16_t tag = 0) : tptr(ptr) { setTag(tag); }

	/**
	 * @brief extract the 48-bit the pointer part
	 *
	 * @return ptr The actuall pointer part of tptr
	 */
	template <typename T>
	[[nodiscard]] inline auto getPtr() const -> T* {
		// ignore the least significant bit of the tagged pointer
		// NOLINTNEXTLINE(performance-no-int-to-ptr)
		return reinterpret_cast<T*>(tptr & (TPTR_POINTER_MASK - 1));
	}

	/**
	 * @brief extract the 48-bit pointer part
	 *
	 * The least significant bit is masked since it is used as marker
	 * and not used by a pointer due to alignment.
	 *
	 * @return ptr The value of the actual pointer part of tptr
	 */
	[[nodiscard]] inline auto getPtrValue() const -> uintptr_t {
		// ignore the least significant bit of the tagged pointer
		// NOLINTNEXTLINE(performance-no-int-to-ptr)
		return tptr & (TPTR_POINTER_MASK - 1);
	}

	/**
	 * @brief extract the 48-bit the pointer part
	 *
	 * @return ptr The value of the 48 least significant bits
	 */
	[[nodiscard]] inline auto getRawPtrValue() const -> uintptr_t {
		// NOLINTNEXTLINE(performance-no-int-to-ptr)
		return tptr & TPTR_POINTER_MASK;
	}

	/**
	 * @brief extract the 16-bit tag part
	 *
	 * @return tag The 16-bit tag stored in tptr
	 */
	[[nodiscard]] inline auto getTag() const -> uint16_t {
		return static_cast<uint16_t>(tptr >> TPTR_POINTER_BITS);
	}
	/**
	 * @brief update the 16-bit counter part
	 *
	 * @param tag The tag which should be stored in the tagged ptr
	 */
	inline void setTag(uint16_t tag) {
		uintptr_t stripped_ptr = tptr & TPTR_POINTER_MASK;
		tptr = stripped_ptr | ((uintptr_t)tag << TPTR_POINTER_BITS);
	}

	/**
	 * @brief check if the least significant bit is set
	 *
	 * @return marked True if the least significant bit is set, False otherwise
	 */
	[[nodiscard]] inline auto isMarked() const -> bool { return (tptr & 1) == 1; }

	/**
	 * @brief set or unset the least significant bit
	 *
	 * @param mark True if the least significant bit should be set, False if it should be unset
	 */
	inline void setMark(bool mark) { tptr = mark ? (tptr | 1) : tptr & (~1); }

	/**
	 * @brief unfold both parts of a tagged pointer
	 *
	 * @param ptr A pointer where the pointer part should be stored
	 * @return tag The 16-bit tag stored in tptr
	 *
	 */
	template <typename T>
	inline auto unfold(T** ptr) const -> uint16_t {
		*ptr = getPtr<T>();
		return getTag();
	}

	inline operator uintptr_t() const { return tptr; }

	// NOLINTNEXTLINE(performance-no-int-to-ptr)
	inline operator void*() const { return reinterpret_cast<void*>(tptr); }

	inline operator bool() const { return tptr != 0; }
};
}	 // namespace emper::lib
