// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#pragma once

#include <pthread.h>

#include <cstring>
#include <queue>

#include "Common.hpp"

namespace lib::adt {

template <typename I>
class RwLockUnboundedQueue {
 private:
	pthread_rwlock_t lock;

	std::queue<I*> queue;

	void acquireWriteLock() {
		int err = pthread_rwlock_wrlock(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_wrlock failed: " << strerror(err));
		}
	}

	void acquireReadLock() {
		int err = pthread_rwlock_rdlock(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_rdlock failed: " << strerror(err));
		}
	}

	void unlock() { pthread_rwlock_unlock(&lock); }

	void upgradeLock() {
		unlock();
		acquireWriteLock();
	}

 public:
	RwLockUnboundedQueue() {
		int err = pthread_rwlock_init(&lock, nullptr);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_init failed: " << strerror(err));
		}
	}

	~RwLockUnboundedQueue() {
		int err = pthread_rwlock_destroy(&lock);
		if (unlikely(err)) {
			DIE_MSG("pthread_rwlock_destroy failed: " << strerror(err));
		}
	}

	void enqueue(I* item) {
		acquireWriteLock();
		queue.push(item);
		unlock();
	}

	template <class InputIt>
	void insert(InputIt begin, InputIt end) {
		acquireWriteLock();
		for (; begin != end; ++begin) {
			queue.push(*begin);
		}
		unlock();
	}

	void insert(I** items, unsigned count) {
		acquireWriteLock();
		for (unsigned i = 0; i < count; ++i) {
			queue.push(items[i]);
		}
		unlock();
	}

	auto dequeue() -> I* {
		I* res = nullptr;

		acquireReadLock();
		if (queue.empty()) {
			goto done;
		}

		upgradeLock();
		if (queue.empty()) {
			goto done;
		}

		res = queue.front();
		queue.pop();

	done:
		unlock();
		return res;
	}

	template <class OutputIt>
	auto dequeue(OutputIt begin, OutputIt end) -> size_t {
		acquireReadLock();
		OutputIt cur = begin;
		if (queue.empty()) {
			goto done;
		}

		upgradeLock();
		if (queue.empty()) {
			goto done;
		}

		for (; !queue.empty() && cur != end; ++cur) {
			*cur = queue.front();
			queue.pop();
		}

	done:
		unlock();
		return cur - begin;
	}
};
}	 // namespace lib::adt
