// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <atomic>
#include <cstdint>

#include "Common.hpp"
#include "StealingResult.hpp"
#include "lib/adt/PushBottomResult.hpp"

namespace adt {

/**
 * @file
 *
 * A CL queue used for work stealing.
 *
 * Implements the partially concurrent queue for the work stealing
 * scheduling algorithm as defined in "Dynamic Circular Work-Stealing Deque" by Chace and Lev. The
 * worker uses pushBottom to push new items on the stack, and
 * retrieves items via popBottom. Eventually, when its own queue is
 * empty, it will use popTop to retrieve items from a different
 * Worker's queue.
 *
 * Only pushBottom/popBottom and popTop may be used concurrently,
 * i.e. concurrent use of pushBottom with popBottom is not allowed.
 *
 * Unlike the ABP queue, the CL queue uses a circular array and
 * only-incrementing uint64_t bottom and top fields. According to the
 * authors "A 64-bit integer is large enough to accommodate 64 years
 * of pushes, pops and steals executing at a rate of 4 billtion
 * operations per second", so overflows should be no problem.
 */
template <typename ITEM_TYPE, const uintptr_t CAPACITY>
class WsClV2Queue {
	// Bottom of the queue is only modified in pushBottom() and
	// popBottom() methods, which are only used by the same
	// thread. Thus we don't need to make it atomic.
	// TODO should this become a std::atomic?
	ALIGN_TO_CACHE_LINE uint64_t bottom{1};

	ALIGN_TO_CACHE_LINE std::atomic<uint64_t> top;

	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	ITEM_TYPE queue[CAPACITY];

 public:
	// 'bottom' and 'top' are initialized to '1', instead of '0' as
	// it's done in the "Dynamic Circular Work-Stealing Deque" paper
	// because popBottom will decrement bottom, which will result in
	// an underflow if bottom is '0'. The paper's queue uses Java
	// 'long' for bottom and top and is thus safe since it's signed.
	WsClV2Queue() : top(1) {}
	auto pushBottom(ITEM_TYPE item) -> PushBottomResult;
	auto popBottom(ITEM_TYPE *item) -> bool;
	template <const int maxRetries>
	auto popTop(ITEM_TYPE *item) -> emper::StealingResult;
	[[nodiscard]] auto isFull() const -> bool;
	[[nodiscard]] auto isEmpty() const -> bool;
	void print() const;
	[[nodiscard]] inline auto capacity() const -> uintptr_t { return CAPACITY; }
	[[nodiscard]] inline auto usedSlots() const -> uint64_t { return bottom - top; }
	[[nodiscard]] inline auto freeSlots() const -> uintptr_t { return capacity() - usedSlots(); }
};

template <typename ITEM_TYPE, const uintptr_t CAPACITY>
auto WsClV2Queue<ITEM_TYPE, CAPACITY>::pushBottom(const ITEM_TYPE item) -> PushBottomResult {
	bool pushed = true;
	// Check if queue is full.
	if (isFull()) {
		pushed = false;
		goto out;
	}

	queue[bottom % CAPACITY] = item;

	// TODO: Possible to move after fence?
	bottom += 1;

	// Write fence / memory barrier
	atomic_thread_fence(std::memory_order_release);

out:
	return PushBottomResult{pushed, bottom - top};
}

template <typename ITEM_TYPE, const uintptr_t CAPACITY>
auto WsClV2Queue<ITEM_TYPE, CAPACITY>::popBottom(ITEM_TYPE *item) -> bool {
	// TODO: new bottom value must be flushed out of the store buffer
	const uint64_t localBottom = --bottom;
	uint64_t localTop = top.load(std::memory_order_acquire);

	if (localTop > localBottom) {
		// The queue is empty.
		bottom = localTop;
		return false;
	}

	*item = queue[localBottom % CAPACITY];

	// There was no possiblity of a race. No atomic CAS required.
	if (localBottom > localTop) return true;

	bool res = top.compare_exchange_weak(localTop, localTop + 1, std::memory_order_release,
																			 std::memory_order_relaxed);
	// Either a popTop() removed the element ('res' is false) or we
	// removed the element ('res' is true), but we need to increment
	// the 'bottom' value, since the element bottom pointed at is now
	// gone. N.B. bottom does point to the next free slot, the actual
	// element we remove is bottom-1.
	bottom = localBottom + 1;
	return res;
}

template <typename ITEM_TYPE, const uintptr_t CAPACITY>
template <const int maxRetries>
auto WsClV2Queue<ITEM_TYPE, CAPACITY>::popTop(ITEM_TYPE *item) -> emper::StealingResult {
	int retries = 0;
	uint64_t localTop = top.load(std::memory_order_acquire);

loop:
	ATTR_UNUSED;
	// Initially this fence was after the check if the queue is
	// empty. But since the shared variable 'bottom' is read by the
	// check, I believe it belongs before the check.
	// Read fence / memory barrier
	atomic_thread_fence(std::memory_order_acquire);

	// Check if the queue is empty.
	if (bottom <= localTop) return emper::StealingResult::Empty;

	*item = queue[localTop % CAPACITY];

	// If cas fails, then this popTop() lost the race against another
	// popTop() or a rare-case popBottom().
	if (top.compare_exchange_weak(localTop, localTop + 1, std::memory_order_release,
																std::memory_order_relaxed)) {
		return emper::StealingResult::Stolen;
	}

	// Loop indefinitely
	if constexpr (maxRetries < 0) goto loop;
	// Loop maxRetries times
	if constexpr (maxRetries > 0) {
		if (retries == maxRetries) return emper::StealingResult::LostRace;

		++retries;
	}

	return emper::StealingResult::LostRace;
}

template <typename ITEM_TYPE, const uintptr_t CAPACITY>
auto WsClV2Queue<ITEM_TYPE, CAPACITY>::isFull() const -> bool {
	return usedSlots() >= CAPACITY;
}

template <typename ITEM_TYPE, const uintptr_t CAPACITY>
auto WsClV2Queue<ITEM_TYPE, CAPACITY>::isEmpty() const -> bool {
	return top >= bottom;
}

}	 // namespace adt
