// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include "sleep_strategy/PipeSleepStrategy.hpp"

#include <liburing.h>

#include <atomic>
#include <cassert>

#include "CallerEnvironment.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "io/IoContext.hpp"
#include "lib/TaggedPtr.hpp"

using emper::io::IoContext;
using emper::lib::TaggedPtr;

namespace emper::sleep_strategy {

thread_local PipeSleepStrategy::Pipe PipeSleepStrategy::pipe;
thread_local PipeSleepStrategy::SleepState PipeSleepStrategy::sleepState;

PipeSleepStrategy::PipeSleepStrategy(Runtime& runtime, workerid_t workerCount)
		: workerCount(workerCount), stats(runtime) {
	LOGD("init pipe sleep startegy");
	sleepers.store(0, std::memory_order_relaxed);

	sleepStates = new SleepState*[workerCount];
	pipes = new Pipe*[workerCount];

	runtime.addNewWorkerHook([this](workerid_t workerId) {
		sleepStates[workerId] = &sleepState;
		pipes[workerId] = &pipe;
	});
}

PipeSleepStrategy::~PipeSleepStrategy() {
	delete[] sleepStates;
	delete[] pipes;
}

void PipeSleepStrategy::sleep() {
	IoContext& io = *IoContext::getWorkerIo();

	{
		std::lock_guard<std::mutex> lock(sleepState.lock);

		// Check if we were notified specifically and should skip the sleep attempt
		if (sleepState.isNotified()) {
			sleepState.markRunning();
			LOGD("Reset notified state to running -> skip sleeping");
			return;
		}

		if (!sleepState.globalReadInflight) {
			// increment the sleeper count if it was negative we should skip sleeping
			int64_t sleeping = sleepers.fetch_add(1, std::memory_order_acquire);
			if (sleeping < 0) {
				LOGD("observed sleeper count as: " << sleeping << " -> skip sleeping");
				return;
			}
			assert(sleeping <= workerCount);

			sleepState.globalReadInflight = true;

			struct io_uring_sqe* sqe = io.getSqe();

			// We read directly into the workers dispatchHint
			io_uring_prep_read(sqe, global.sleepFd, &io.worker->dispatchHint,
												 sizeof(io.worker->dispatchHint), 0);

			// Mark the sqe as a new work notification to reset the Global flag when reaping the
			// resulting cqe
			io_uring_sqe_set_data(
					sqe, TaggedPtr(0, static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification)));

			io.trackReqsInUring(1);

			LOGD("prepared global.sleepFd read and set sleepers count to: " << sleeping + 1);
		}

		if (!sleepState.isSleeping()) {
			sleepState.markSleeping();

			struct io_uring_sqe* sqe = io.getSqe();

			// We read directly into the workers dispatchHint
			// TODO: Think about the race between the two reads
			io_uring_prep_read(sqe, pipe.sleepFd, &io.worker->dispatchHint,
												 sizeof(io.worker->dispatchHint), 0);

			// Tag the sqe with a marked new work notification to reset the specific state
			// when reaping the resulting cqe.
			io_uring_sqe_set_data(
					sqe, TaggedPtr((void*)nullptr,
												 static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification), true));

			io.trackReqsInUring(1);
			LOGD("prepared pipe.sleepFd read");
		}
	}

	// Wait for IO completions
	LOGD("sleep until IO completions occur");
	io.submitPreparedSqesAndWait<CallerEnvironment::EMPER>(1);
	LOGD("wakeup due to available IO completions");

	stats.incWakeup();
}

template <CallerEnvironment callerEnvironment>
void PipeSleepStrategy::notifySpecific(workerid_t workerId) {
	auto& specificState = *sleepStates[workerId];
	std::lock_guard<std::mutex> lock(specificState.lock);

	LOGD("Specifically notify  worker " << workerId << " from " << callerEnvironment);

	if (specificState.isNotified()) {
		LOGD(workerId << " already marked notified");
		return;
	}

	const bool isSleeping = specificState.markNotified();

	if (isSleeping) {
		LOGD(workerId << " has specific read -> write notification");
		writeSpecificNotification(workerId);
	}
}

template void PipeSleepStrategy::notifySpecific<CallerEnvironment::EMPER>(workerid_t workerId);
template void PipeSleepStrategy::notifySpecific<CallerEnvironment::ANYWHERE>(workerid_t workerId);

template <CallerEnvironment callerEnvironment>
void PipeSleepStrategy::onNewWorkNotification(IoContext& io, TaggedPtr data) {
	if constexpr (callerEnvironment != CallerEnvironment::OWNER) {
		DIE_MSG("Others reaping new work notification from " << io.worker->getWorkerId() << " CQ");
	}

	assert(data.getTag() == static_cast<uint16_t>(IoContext::PointerTags::NewWorkNotification));

	if (data.isMarked()) {
		std::lock_guard<std::mutex> lock(sleepState.lock);
		LOGD("Got specific notification");
		stats.incWakeupDueToNotify();

		// Reset specific and notified flag to indicate that a new specific notification
		// was consumed, a new specific read must be prepared and other notifySpecific
		// calls must notify again.
		sleepState.markRunning();
	} else {
		LOGD("Got new work notification");
		stats.incWakeupDueToNotify();
		// Reset global flag to indicate that a new sleep cqe must be prepared
		// and allow the completer to reap completions again
		sleepState.globalReadInflight = false;
	}
}

template void PipeSleepStrategy::onNewWorkNotification<CallerEnvironment::OWNER>(IoContext& io,
																																								 TaggedPtr data);
template void PipeSleepStrategy::onNewWorkNotification<CallerEnvironment::EMPER>(IoContext& io,
																																								 TaggedPtr data);
template void PipeSleepStrategy::onNewWorkNotification<CallerEnvironment::ANYWHERE>(IoContext& io,
																																										TaggedPtr data);
}	 // namespace emper::sleep_strategy
