// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <unistd.h>

#include <algorithm>
#include <atomic>
#include <cstdint>
#include <iostream>
#include <mutex>
#include <vector>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "FiberHint.hpp"
#include "emper-common.h"
#include "lib/TaggedPtr.hpp"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"
#include "sleep_strategy/SleeperState.hpp"
#include "sleep_strategy/Stats.hpp"

class Runtime;

namespace emper::io {
class IoContext;
}

namespace emper::sleep_strategy {
/**
 * @brief A pipe(2) based sleep strategy combining the IO subsystem with the sleep strategy
 *
 * implement a pipe based sleep strategy using the IO subsystem
 *
 * Design goals
 * ============
 *
 * * Wakeup either on external newWork notifications or on local IO completions
 *   -> Sleep strategy is sound without the IO completer
 * * Do as less as possible in a system saturated with work
 * * Pass a hint where to find new work to suspended workers
 *
 * Algorithm
 * =========
 *
 *```
 * Data:
 *     Global:
 *             global pipe
 *             sleepers count
 *     Per worker:
 *             dispatch hint buffer
 *             specific pipe
 *             state lock
 *             sleep state
 *
 * Sleep:
 *     Lock state
 *     If state == notified
 *         Set state = running
 *         return
 *
 *     If we have no global sleep request in flight
 *             Atomic increment sleep count
 *             Skip sleeping if sleep count was < 0
 *             Mark sleep requests in flight
 *             Prepare read cqe from the global pipe to dispatch hint buffer
 *
 *     If state == running
 *             Set state = sleeping
 *             Prepare marked read cqe from the specific pipe to dispatch hint buffer
 *
 *     Unlock state
 *     Wait until IO completions occurred
 *
 * NotifyEmper(n):
 *    if observed sleepers <= 0
 *            return
 *
 *    // Determine how many we are responsible to wake
 *    do
 *            toWakeup = min(observed sleepers, n)
 *    while (!CAS(sleepers, toWakeup))
 *
 *    write toWakeup hints to the hint pipe
 *
 * NotifyAnywhere(n):
 *    // Ensure all n notifications take effect
 *    while (!CAS(sleepers, observed sleepers - n))
 *        if observed sleeping <= -n
 *            return
 *
 *    toWakeup = min(observed sleeping, n)
 *    write toWakeup hints to the hint pipe
 *
 * NotifySpecific(w):
 *    Get w's state
 *    Lock state
 *    Return if already notified
 *    Mark state notified
 *    Write hint if was stat was sleeping
 *
 * onNewWorkCompletion:
 *     If data is marked
 *         lock state
 *         set state = running
 *         return
 *
 *     Reset global read inflight
 *```
 *
 * Notes
 * =====
 *
 * * We must decrement the sleepers count on the notifier side to
 *   prevent multiple notifiers to observe all the same amount of sleepers,
 *   trying to wake up the same sleepers by writing to the pipe and jamming it up
 *   with unconsumed hints and thus blocking in the notify write resulting
 *   in a deadlock.
 * * The CAS loops on the notifier side are needed because decrementing
 *   and incrementing the excess is racy: Two notifier can observe the
 *   sum of both their excess decrement and increment to much resulting in a
 *   broken counter.
 * * Add the dispatch hint code in AbstractWorkStealingScheduler::nextFiber.
 *   This allows workers to check the dispatch hint after there
 *   where no local work to execute.
 *   This is a trade-off where we trade slower wakeup - a just awoken worker
 *   will check for local work - against a faster dispatch hot path when
 *   we have work to do in our local WSQ.
 * * Other thread must not reap new work notifications because this
 *   would introduces a race for cqes and a possible
 *   lost wakeup if the other consumes the completions before the worker
 *   is actually waiting for them.
 * * When notifying sleeping workers from anywhere we must ensure that all
 *   notifications take effect. This is needed for example when terminating
 *   the runtime to prevent sleep attempt from worker thread which are
 *   about to sleep but have not incremented the sleeper count yet.
 *   We achieve this by always decrementing the sleeper count by the notification
 *   count.
 */
class PipeSleepStrategy : AbstractWorkerSleepStrategy<PipeSleepStrategy>,
													public Logger<LogSubsystem::SLEEP_S> {
	friend class emper::io::IoContext;

	/**
	 * @brief State of a worker
	 */
	class SleepState {
		friend class PipeSleepStrategy;

		bool globalReadInflight = false;
		std::mutex lock;
		emper::sleep_strategy::SleeperState s = emper::sleep_strategy::SleeperState::Running;

		auto markNotified() -> bool {
			auto oldS = s;
			s = emper::sleep_strategy::SleeperState::Notified;
			return oldS == emper::sleep_strategy::SleeperState::Sleeping;
		}
		void markSleeping() { s = emper::sleep_strategy::SleeperState::Sleeping; }
		void markRunning() { s = emper::sleep_strategy::SleeperState::Running; }

		auto isNotified() const -> bool { return s == emper::sleep_strategy::SleeperState::Notified; }
		auto isSleeping() const -> bool { return s == emper::sleep_strategy::SleeperState::Sleeping; }
	};

	SleepState** sleepStates;
	static thread_local SleepState sleepState;

	class Pipe {
		friend class PipeSleepStrategy;
		int sleepFd;
		int notifyFd;

		Pipe() {
			// NOLINTNEXTLINE(modernize-avoid-c-arrays)
			int fds[2];
			if (::pipe(fds)) DIE_MSG_ERRNO("pipe failed");
			sleepFd = fds[0];
			notifyFd = fds[1];
		}

		~Pipe() {
			if (close(sleepFd) || close(notifyFd)) DIE_MSG("Closing pipe failed");
		}
	};

	const workerid_t workerCount;
	Pipe global;

	/**
	 * @brief Per worker pipe to notify specific workers
	 */
	Pipe** pipes;
	static thread_local Pipe pipe;

	Stats stats;

	// Make sure the shared counter lives in an exlusive cache line
	CACHE_LINE_EXCLUSIVE(std::atomic<int64_t>, sleepers);

	void writeNotifications(emper::FiberHint hint, int64_t count) {
		stats.addNotifications(count);

		std::vector<void*> hints(count, hint);

		ssize_t res = write(global.notifyFd, hints.data(), sizeof(void*) * hints.size());
		if (unlikely(res) < 0) {
			DIE_MSG_ERRNO("writing to notifyFd failed");
		}
	}

	void writeSpecificNotification(workerid_t workerId) {
		stats.addNotifications(1);

		const void* b = nullptr;
		ssize_t res = write(pipes[workerId]->notifyFd, &b, sizeof(b));
		if (unlikely(res) < 0) {
			DIE_MSG_ERRNO("writing to the notifyFd of worker " << workerId << " failed");
		}
	}

	void notifyFromEmper(int64_t& sleeping, emper::FiberHint hint, int64_t count) {
		int64_t toWakeup;

		// We need to decrement the sleeper count on the notification side
		// to prevent overcommiting notifications, filling up the pipe which results
		// in notify calls to block on write.
		// And the CAS loop is needed because we don't want to decrement the sleepers
		// count to much.

		// Decrement the sleeper count for each sleeper we are going to wakeup
		do {
			// No one is sleeping -> we don't need to notify
			if (sleeping <= 0) {
				return;
			}

			// Count of sleepers we notify
			toWakeup = std::min(sleeping, count);
		} while (!sleepers.compare_exchange_weak(sleeping, sleeping - toWakeup,
																						 std::memory_order_release, std::memory_order_acquire));

		// increment the notify stat if we actually have to notify someone
		stats.incNotify();
		writeNotifications(hint, toWakeup);

		LOGD("notifyFromEmper written "
				 << toWakeup << " notifications and set sleepers count to: " << sleeping - toWakeup);
	}

	void notifyFromAnywhere(int64_t& sleeping, emper::FiberHint hint, int64_t count) {
		// If we observe nobody sleeping we need to prevent sleep locks when
		// everybody is about to sleep but none has incremented the sleepers count yet.
		// We prevent the next sleep from blocking by setting the sleeper count to
		// the negative amount we want to prevent from sleeping.

		// More general if we notify from anywhere we need all notifications to take
		// effect. Only notifying the sleepers we observe may not be enough.
		// For example when terminating the runtime we notify all workers if we
		// only wakeup the ones we observe sleeping we may never terminate because
		// after we notified all we observed new one went to sleep which will never
		// be notified.
		// To prevent this we make sure that all count notifications take effect
		// by making the sleepers count negative and thus preventing at least count
		// sleep attempts.

		int64_t newSleeping;
		do {
			// We already prevent enough sleep attempts
			if (sleeping <= -count) {
				LOGD("notifyFromAnywhere sleeper count already preventing enough sleep attempts");
				return;
			}
			// Don't decrease the sleeper count in the CAS loop further than -count,
			// which is the threshold we need to ensure that the notifications will be
			// observed.
			// Decreasing it further than this threshold is not faulty it just
			// results in unnecessary skipped sleeps.
			newSleeping = std::max(-count, sleeping - count);
		} while (!sleepers.compare_exchange_weak(sleeping, newSleeping, std::memory_order_release,
																						 std::memory_order_acquire));

		stats.incNotify();

		// To wakeup should not be negative but could be if sleeping was already negative
		int64_t toWakeup = std::max(std::min(sleeping, count), 0L);
		if (toWakeup > 0) {
			writeNotifications(hint, toWakeup);
		}

		LOGD("notifyFromAnywhere written "
				 << toWakeup << " notifications and set sleepers count to: " << newSleeping);
	}

	template <CallerEnvironment callerEnvironment>
	void notify(emper::FiberHint hint, uint32_t count) {
		// The hint must be != nullptr so sleep() knows when to prepare and submit
		// a sleepFd read sqe.

		stats.incOnNewWork();

		int64_t sleeping = getSleeping();
		auto signedCount = static_cast<int64_t>(count);

		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			notifyFromAnywhere(sleeping, hint, signedCount);
		} else {
			notifyFromEmper(sleeping, hint, signedCount);
		}
	}

 public:
	PipeSleepStrategy(Runtime& runtime, workerid_t workerCount);
	~PipeSleepStrategy();

	void printStats(std::ostream& out = std::cout) { stats.print(out); }

	[[nodiscard]] inline auto getSleeping() const -> long { return sleepers.load(); }

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		notifyMany<callerEnvironment>(1);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (count > 1) LOGD("notify(hint, " << count << ")");
		notify<callerEnvironment>(emper::FiberHint::createNewWorkHint<callerEnvironment>(), count);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		notifyMany<callerEnvironment>(workerCount);
	}

	template <CallerEnvironment callerEnvironment>
	void notifySpecific(workerid_t workerId);

	template <CallerEnvironment callerEnvironment>
	void onNewWorkNotification(emper::io::IoContext& io, emper::lib::TaggedPtr data);
	void sleep();
};
}	 // namespace emper::sleep_strategy
