// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Schmaus, Florian Fischer
#include "AbstractWorkStealingScheduler.hpp"

#include <algorithm>
#include <array>
#include <cassert>
#include <cstdint>
#include <ostream>	// for operator<<, basic_ostream<>::__ostream_type
#include <vector>

#include "AbstractFiber.hpp"
#include "CallerEnvironment.hpp"
#include "Common.hpp"	 // for unlikely, likely
#include "Debug.hpp"	 // for ABORT
#include "Emper.hpp"	 // for OVERFLOW_QUEUE
#include "Fiber.hpp"
#include "FiberHint.hpp"
#include "FiberSource.hpp"
#include "NextFiberResult.hpp"
#include "Runtime.hpp"	// for Runtime
#include "StealingResult.hpp"
#include "Worker.hpp"
#include "emper-common.h"	 // for workerid_t
#include "io/IoContext.hpp"
#include "lib/adt/PushBottomResult.hpp"
#include "strategies/AbstractWorkStealingStrategy.hpp"
#include "strategies/AbstractWorkStealingWorkerStats.hpp"

using awss = AbstractWorkStealingStrategy;

using emper::StealingResult;
using emper::io::IoContext;

thread_local AbstractWorkStealingScheduler::WsQueue<AbstractWorkStealingScheduler::QUEUE_SIZE>
		AbstractWorkStealingScheduler::queue;

thread_local AbstractWorkStealingScheduler::MpscQueue AbstractWorkStealingScheduler::mpscQueue;

AbstractWorkStealingScheduler::AbstractWorkStealingScheduler(Runtime& runtime,
																														 RuntimeStrategy& strategy)
		: Scheduler(runtime, strategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new AbstractWorkStealingScheduler::WsQueue<QUEUE_SIZE>*[workerCount];
	mpscQueues = new AbstractWorkStealingScheduler::MpscQueue*[workerCount];

	auto newWorkerHook = [this](workerid_t workerId) {
		queues[workerId] = &queue;
		mpscQueues[workerId] = &mpscQueue;
	};
	addNewWorkerHook(newWorkerHook);
}

void AbstractWorkStealingScheduler::scheduleViaWorkStealing(Fiber& fiber) {
	auto pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
		if constexpr (emper::OVERFLOW_QUEUE) {
			enqueueInAnywhereQueue(fiber);

			emper::statsIncr(awss::stats.scheduledFibersToOverflowQueue);
		} else {
			ABORT("Could not push fiber " << &fiber << " into queue");
		}
	} else if constexpr (emper::STATS) {
		uint64_t queueLength = pushed.size;
		awss::stats.recordScheduledToLocalAndQueueLength(queueLength);
	}

	// Classes using this method are supposed to always invoke this
	// method. Hence we call onNewWork() here.
	onNewWork<CallerEnvironment::EMPER>(emper::FiberHint{emper::FiberSource::local});
}

void AbstractWorkStealingScheduler::scheduleToMpscQueue(Fiber& fiber, workerid_t workerId) {
	mpscQueues[workerId]->enqueue(&fiber);
	emper::statsIncr(awss::stats.scheduledFibersToMpscQueue);

	// Notify the runtime that we scheduled work to a specific workers' mpscQueue
	// the runtime is responsible to notify it if necessary.
	onNewWork<CallerEnvironment::EMPER>(emper::FiberHint{workerId, emper::FiberSource::mpscQueue});
}

auto AbstractWorkStealingScheduler::maybeRecycle(AbstractFiber* fiber) -> bool {
	if (fiber->isRunnable()) return false;

	recycle(fiber);
	return true;
}

auto AbstractWorkStealingScheduler::nextFiberViaAnywhereQueue() -> std::optional<NextFiberResult> {
	const size_t KEEP_FREE_SLOTS = 64;
	const size_t DEQUEUE_FROM_ANYWHERE_MAX = 512;

	const uintptr_t remainingCapacityLocalWsQueue = queue.freeSlots();
	size_t maxFiberCount;
	if (remainingCapacityLocalWsQueue > KEEP_FREE_SLOTS) {
		maxFiberCount = remainingCapacityLocalWsQueue - KEEP_FREE_SLOTS;
		maxFiberCount = std::min(maxFiberCount, DEQUEUE_FROM_ANYWHERE_MAX);
	} else {
		maxFiberCount = 1;
	}

	std::array<Fiber*, DEQUEUE_FROM_ANYWHERE_MAX> fibers;

	size_t fiberCount =
			dequeueFibersFromAnywhereQueue(fibers.begin(), fibers.begin() + maxFiberCount);

	Fiber* res = nullptr;
	size_t fibersLiftedFromAnywhere = 0;
	for (size_t i = 0; i < fiberCount; i++) {
		Fiber* fiber = fibers[i];

		if (maybeRecycle(fiber)) continue;

		// We found a runnable fiber. First check if we already have one
		// fiber saved to be returned.
		if (!res) {
			res = fiber;
			continue;
		}

		// We already have a fiber saved which will be returned. Hence we
		// push the remaining runnable fibers to our work-stealing queue,
		// so that they can be stolen by others.
		[[maybe_unused]] bool pushed = queue.pushBottom(fiber);

		// We have ensured that there is capacity for this fiber, hence we
		// must be able to push it here.
		assert(pushed);

		fibersLiftedFromAnywhere++;
	}

	if (fibersLiftedFromAnywhere) {
		onNewWork<CallerEnvironment::EMPER>(emper::FiberHint{emper::FiberSource::local});

		if constexpr (emper::STATS) {
			awss::stats.recordFibersLiftedFromAnywhereQueue(fibersLiftedFromAnywhere);
		}
	}

	if (res) return NextFiberResult{res, emper::FiberSource::anywhereQueue};

	return std::nullopt;
}

auto AbstractWorkStealingScheduler::tryStealFiberFrom(workerid_t victim)
		-> std::optional<NextFiberResult> {
	constexpr int maxRetries = emper::WAITFREE_WORK_STEALING ? 0 : -1;
	AbstractFiber* fiber;
popTop:
	StealingResult res = queues[victim]->popTop<maxRetries>(&fiber);
	if (res == StealingResult::Stolen) {
		emper::statsIncr(awss::stats.nextFiberStolen);

		if (maybeRecycle(fiber)) goto popTop;

		return NextFiberResult{fiber, emper::FiberSource::stolen};
	}

	if constexpr (emper::IO_STEALING) {
		auto* victimIo = runtime.ioContexts[victim];
		fiber = victimIo->reapSingleCompletion<CallerEnvironment::EMPER>();
		if (fiber) {
			emper::statsIncr(awss::stats.nextIoFiberStolen);
			return NextFiberResult{fiber, emper::FiberSource::ioStolen};
		}
	}

	return std::nullopt;
}

auto AbstractWorkStealingScheduler::nextFiberResultViaWorkStealing()
		-> std::optional<NextFiberResult> {
	AbstractFiber* fiber;

popBottom:
	bool poped = queue.popBottom(&fiber);
	if (likely(poped)) {
		emper::statsIncr(awss::stats.nextFiberFromLocal);

		if (maybeRecycle(fiber)) goto popBottom;

		return NextFiberResult{fiber, emper::FiberSource::local};
	}

	auto* const currentWorker = Worker::getCurrentWorker();

	if constexpr (emper::WORKER_IGNORE_WAKEUP_HINT) {
		goto work_stealing;
	}

	// Try dispatch hint possibly set by the sleep strategy
	if (currentWorker->dispatchHint) {
		const emper::FiberHint dispatchHint = currentWorker->dispatchHint;
		currentWorker->dispatchHint.clear();
		const auto source = dispatchHint.getSource();

		switch (source) {
			case emper::FiberSource::hintWsq: {
				const auto victim = dispatchHint.getWorker();
				const auto stolen = tryStealFiberFrom(victim);
				if (stolen) {
					emper::statsIncr(awss::stats.nextFiberFromHintLocal);
					onWorkStolen();
					return NextFiberResult{(*stolen).fiber, emper::FiberSource::hintWsq};
				}
			} break;

			case emper::FiberSource::hintAq: {
				const auto fromAnywhere = nextFiberViaAnywhereQueue();
				if (fromAnywhere) {
					emper::statsIncr(awss::stats.nextFiberFromHintAnywhere);
					return NextFiberResult{(*fromAnywhere).fiber, emper::FiberSource::hintAq};
				}
				break;
			}
			default:
				DIE_MSG("invalid dispatch hint");
		}
	}

work_stealing:
	// Go into work stealing
	// TODO: Determine if there is a better value than 1/3.
	const float CHECK_ANYWHERE_QUEUE_AT_PERCENTAGE = 0.33;
	const workerid_t myWorkerId = currentWorker->getWorkerId();
	const workerid_t workerCount = runtime.getWorkerCount();
	// NOLINTNEXTLINE(bugprone-narrowing-conversions)
	const workerid_t checkAnywhereQueueAt = workerCount * CHECK_ANYWHERE_QUEUE_AT_PERCENTAGE;

	workerid_t startWorkerId = currentWorker->nextRandomWorkerId();
	// TODO: See how changing the victim count affects things.
	const workerid_t victimCount = [&] {
		if constexpr (emper::WS_VICTIM_COUNT)
			return emper::WS_VICTIM_COUNT;
		else if constexpr (emper::WS_VICTIM_DENOMINATOR)
			return workerCount / emper::WS_VICTIM_DENOMINATOR;
		else
			return workerCount;
	}();

	for (workerid_t i = 0; i < victimCount; ++i) {
		workerid_t victim = (startWorkerId + i) % workerCount;

		// Don't steal from ourselves.
		if (unlikely(victim == myWorkerId)) continue;

		auto stolenFiber = tryStealFiberFrom(victim);
		if (stolenFiber) {
			onWorkStolen();
			return *stolenFiber;
		}

		if constexpr (!emper::CHECK_ANYWHERE_QUEUE_WHILE_STEALING) continue;

		// If we failed to steal from a certain number of victims, check
		// the anywhere queue for new fibers.
		if (i == checkAnywhereQueueAt) {
			auto anywhereQueueFiber = nextFiberViaAnywhereQueue();
			if (anywhereQueueFiber) return *anywhereQueueFiber;
		}
	}

	// Try the "scheduled from anywhere" queue to get work as last resort.
	auto anywhereQueueFiber = nextFiberViaAnywhereQueue();
	if (anywhereQueueFiber) return *anywhereQueueFiber;

	// We where not able to dequeue any fiber if we reach this point.
	return std::nullopt;
}

auto AbstractWorkStealingScheduler::nextFiberResultFromMpscQueue()
		-> std::optional<NextFiberResult> {
	Fiber* fiber = mpscQueue.dequeue();
	if (fiber != nullptr) return NextFiberResult{fiber, emper::FiberSource::mpscQueue};

	return std::nullopt;
}

auto AbstractWorkStealingScheduler::nextFiberResultFromMpscQueueOrWorkStealing()
		-> std::optional<NextFiberResult> {
	auto result = nextFiberResultFromMpscQueue();
	if (result) return result;

	return nextFiberResultViaWorkStealing();
}

auto AbstractWorkStealingScheduler::pushBottom(AbstractFiber& fiber) -> bool {
	auto pushed = queue.pushBottom(&fiber);
	if (pushed) [[likely]] {
		onNewWork<CallerEnvironment::EMPER>(emper::FiberHint{emper::FiberSource::local});
		if constexpr (emper::STATS) {
			uint64_t queueLength = pushed.size;
			awss::stats.recordScheduledToLocalAndQueueLength(queueLength);
		}
	}
	return pushed;
}
