// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "LawsDispatcher.hpp"

#include <optional>

#include "Common.hpp"	 // for DIE_MSG
#include "Emper.hpp"
#include "Fiber.hpp"
#include "FiberSource.hpp"
#include "LawsStrategy.hpp"	 // for LawsStrategy
#include "NextFiberResult.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "strategies/laws/LawsWorkerStats.hpp"

class AbstractFiber;

void LawsDispatcher::recycle(Fiber* fiber) {
	// If the ref count has not reached zero yet, do not recycle the
	// fiber. But only if the fiber is a multi fiber, i.e. was placed
	// in multiple queues in the first place, and hence needed to be
	// refcounted.
	if (fiber->isMultiFiber() && decreaseRefCount(fiber)) return;

	Dispatcher::recycle(fiber);
}

void LawsDispatcher::dispatchLoop() {
	while (true) {
		std::optional<NextFiberResult> next = runtime.nextFiber();
		if (!next) {
			dispatchLoopDoSleep();

			continue;
		}

		AbstractFiber* fiber = next->fiber;

		// The isRunnable() method performes an atomic swap on a boolean,
		// which was initialized to true, in order to check if this fiber
		// is runnable.
		if (isRunnable(fiber)) {
			if constexpr (emper::STATS) {
				switch (next->source) {
					case emper::FiberSource::mpscQueue:
						LawsStrategy::stats.dispatchedFibersFromPriority++;
						break;
					case emper::FiberSource::local:
						LawsStrategy::stats.dispatchedFibersFromLocal++;
						break;
					case emper::FiberSource::hintWsq:
						LawsStrategy::stats.dispatchedFibersFromHintLocal++;
						break;
					case emper::FiberSource::hintAq:
						LawsStrategy::stats.dispatchedFibersFromHintAnywhere++;
						break;
					case emper::FiberSource::stolen:
						LawsStrategy::stats.dispatchedFibersStolen++;
						break;
					case emper::FiberSource::anywhereQueue:
						LawsStrategy::stats.dispatchedFibersFromAnywhereQueue++;
						break;
					default:
						DIE_MSG("Unknown fiber source: " << next->source);
						break;
				}
			}

			// Update the affinity if one was set. Note that we do this
			// before dispatching the fiber, as the fiber will change the
			// affinity towards the current worker as soon as it starts
			// executing, i.e. right now. Also, future code may delete the
			// affinity buffer as part of the Fiber's run() method, so it
			// may not longer be valid after dispatching it.
			workeraffinity_t* const affinity = getAffinityBuffer(fiber);
			if (affinity) *affinity = Runtime::getWorkerId();

			// The fiber was marked das runnable. Run it now.
			dispatch(fiber);
		}

		Dispatcher::recycle(fiber);
	}
}
